@@ -1,4 +1,4 @@
-#!/usr/bin/env python
+#!/usr/bin/env python3
 # -*- coding: utf-8 -*-
 
 """
@@ -6,7 +6,7 @@
 
     SDL is the Specification and Description Language (Z100 standard from ITU)
 
-    Copyright (c) 2012-2013 European Space Agency
+    Copyright (c) 2012-2019 European Space Agency
 
     Designed and implemented by Maxime Perrotin
 
@@ -15,16 +15,15 @@
     This module is managing the Copy and Paste functions.
 """
 
+import os
 import traceback
 import logging
 from itertools import chain
-import PySide
+from PySide2.QtCore import *
+from PySide2.QtGui import *
+from PySide2.QtWidgets import *
 
-import ogAST
-import sdlSymbols
-import genericSymbols
-import Renderer
-import Pr
+from . import ogAST, ogParser, sdlSymbols, genericSymbols, Renderer, Pr
 
 __all__ = ['copy', 'paste']
 
@@ -36,6 +35,9 @@ COPY_PASTE = []
 # Actual scene clipboard
 CLIPBOARD = None
 
+# System clipboard used to copy-paste between instances of Opengeode
+# Value is set when Opengeode is initialized
+SYS_CLIPBOARD = None
 
 def copy(selection):
     ''' Create a copy (duplicate) of the selected symbols in AST form '''
@@ -82,7 +84,9 @@ def copy_branch(top_level_item):
     if not isinstance(top_level_item, genericSymbols.HorizontalSymbol):
         next_aligned = top_level_item.next_aligned_symbol()
         while next_aligned and next_aligned.grabber.isSelected():
-            next_ast, next_terminators = next_aligned.get_ast()
+            next_pr_text = '\n'.join(Pr.generate(next_aligned, cpy=True,
+                                            nextstate=False, recursive=True))
+            next_ast, next_terminators = next_aligned.get_ast(next_pr_text)
             terminators.extend(next_terminators)
             branch.append(next_ast)
             next_aligned = next_aligned.next_aligned_symbol()
@@ -91,15 +95,24 @@ def copy_branch(top_level_item):
     res_terminators = terminators
     for term in terminators:
         # Get symbol at terminator coordinates
-        symbols = top_level_item.scene().items(PySide.QtCore.QRectF
+        symbols = top_level_item.scene().items(QRectF
                 (term.pos_x, term.pos_y, term.width, term.height).center())
         for symbol in symbols:
             if (isinstance(symbol, sdlSymbols.State) and [c for c in
                 symbol.childSymbols() if isinstance(c, (sdlSymbols.Input,
-                    sdlSymbols.Connect))]):
+                    sdlSymbols.Connect, sdlSymbols.ContinuousSignal))]):
                 term_branch, term_inators = copy_branch(symbol)
                 branch.extend(term_branch)
                 res_terminators.extend(term_inators)
+
+    if SYS_CLIPBOARD is not None:
+        #  Basic copy of a single branch to the system clipboard
+        ident = top_level_item.common_name \
+                if not isinstance(top_level_item, sdlSymbols.State) \
+                else 'state'
+        SYS_CLIPBOARD.setText("OG_SDL@-@{}@-@{}@-@{}".format(os.getpid(),
+                                                         ident,
+                                                         pr_text))
     return branch, res_terminators
 
 
@@ -107,11 +120,43 @@ def paste(parent, scene):
     '''
         Paste previously copied symbols at selection point
     '''
+    remove_after_paste = False
+    if SYS_CLIPBOARD is not None:
+        #  Get from clipboard only if it comes from another process
+        #  otherwise use the local clipboard which contains more data
+        #  (for the moment the shared clipboard only stores top-level items)
+        shared = SYS_CLIPBOARD.text().split('@-@') or []
+        if len(shared) >= 4 \
+                and shared[0] == "OG_SDL" \
+                and shared[1] != str(os.getpid()):
+            common_name = shared[2]
+            pr_text     = shared[3]
+            #  Copy to the local clipboard
+            ast, _, _, _, terminators = \
+                ogParser.parseSingleElement(common_name, pr_text)
+            COPY_PASTE.append(([ast], terminators))
+            remove_after_paste = True
+
     CLIPBOARD.clear()
     if not parent:
         new_symbols = paste_floating_objects(scene)
+        for each in new_symbols:
+            #  Make sure nested scenes are set properly
+            each.double_click()
     else:
         new_symbols = paste_below_item(parent, scene)
+        def rec_double_click(next):
+            if next is not None:
+                next.double_click()
+                rec_double_click(next.next_aligned_symbol())
+        for each in new_symbols:
+            #  Make sure nested scenes are set properly
+            rec_double_click(each)
+
+    if remove_after_paste:
+        #  Remove from local clipboard if it came from system clipboard
+        #  otherwise it would be added over and over
+        COPY_PASTE.pop()
     return new_symbols
 
 
@@ -129,6 +174,11 @@ def paste_floating_objects(scene):
         procedures = (i for i in item_list if isinstance(i, ogAST.Procedure))
         processes = (i for i in item_list if isinstance(i, ogAST.Process))
         for state in states:
+            # nesting is not copied, because the nested content is a scene
+            # that is just referenced.. content must be actually copied
+            # to a new scene only if a state is renamed (this will preserve
+            # the nesting content of the original state)
+
             # First check if state has already been pasted
             try:
                 new_item = Renderer.render(state, scene=CLIPBOARD,
@@ -151,7 +201,7 @@ def paste_floating_objects(scene):
             new_item = Renderer.render(each, scene, states=states)
             symbols.append(new_item)
             new_item.nested_scene = scene.create_subscene(
-                                               type(new_item).__name__.lower())
+                                                        new_item.context_name)
             # Render recursively, creating any required scene
             try:
                 new_item.nested_scene.render_everything(each.content)
@@ -185,9 +235,9 @@ def paste_below_item(parent, scene):
             if (type(new_item).__name__ in parent.allowed_followers):
                 # Move the item from the clipboard to the scene
                 Renderer.add_to_scene(new_item, scene)
-                new_item.setPos(0, 0)
+                new_item.pos_x = new_item.pos_y = 0.0
                 symbols.append(new_item)
             else:
-                raise TypeError('Cannot paste here ({t1} cannot follow {t2}'
+                raise TypeError('Cannot paste here ({t1} cannot follow {t2})'
                                 .format(t1=type(new_item), t2=type(parent)))
     return symbols
