@@ -8,7 +8,7 @@
     The "Connection" class is the mother class, all other connectors must
     inherit from it and possibly redefine some functions or shape.
 
-    Copyright (c) 2012-2014 European Space Agency
+    Copyright (c) 2012-2020 European Space Agency
 
     Designed and implemented by Maxime Perrotin
 
@@ -18,20 +18,17 @@
 import math
 import logging
 
-from PySide.QtCore import Qt, QPointF, QLineF, Slot
-
-from PySide.QtGui import QGraphicsPathItem, QPainterPath, QGraphicsItem, QPen,\
-                         QPainter, QFont, QGraphicsTextItem, QColor, \
-                         QFontMetrics
-
-import ogParser
-from TextInteraction import EditableText
+from PySide2.QtCore import *
+from PySide2.QtGui import *
+from PySide2.QtWidgets import *
+from . import ogParser
+from .TextInteraction import EditableText
 
 LOG = logging.getLogger(__name__)
 
 
 # pylint: disable=R0904
-class Connection(QGraphicsPathItem, object):
+class Connection(QGraphicsPathItem):
     ''' Connection between two symbols (top-level class) '''
     # Default connections are not selectable
     default_cursor = Qt.ArrowCursor
@@ -53,6 +50,15 @@ class Connection(QGraphicsPathItem, object):
         self.childRect = child.sceneBoundingRect()
         # Activate cache mode to boost rendering by calling paint less often
         self.setCacheMode(QGraphicsItem.DeviceCoordinateCache)
+        # When the child moves, the connection may need to adjust the end point
+        self.child.moved.connect(self.child_moved)
+        # Syntax error indicator
+        self.syntax_error: Boolean = False
+
+    @Slot(float, float)
+    def child_moved(self, delta_x, delta_y):
+        ''' When the connection child moves - redefine in subclasses '''
+        pass
 
     @property
     def start_point(self):
@@ -69,6 +75,11 @@ class Connection(QGraphicsPathItem, object):
         ''' Compute connection intermediate points - redefine in subclasses '''
         return self._middle_points
 
+    def top_level(self):
+        ''' In case the connection has a text, implement top_level()
+            as needed by TextInteraction '''
+        return self.parent
+
     def simple_arrow(self, origin='head', path=None):
         ''' Compute the two points of an vertical arrow head '''
         if origin == 'head':
@@ -78,6 +89,10 @@ class Connection(QGraphicsPathItem, object):
         return (QPointF(endp.x() - 5, endp.y() - 5),
                 QPointF(endp.x() + 5, endp.y() - 5))
 
+    def select(self, selected=True):
+        ''' Select a connection - for future use? '''
+        return
+
     def angle_arrow(self, path, origin='head'):
         ''' Compute the two points of the arrow head with the right angle '''
         if origin == 'tail':
@@ -265,52 +280,52 @@ class SignalList(EditableText):
 class Signalroute(Connection):
     ''' Subclass of Connection used to draw channels between processes '''
     in_sig = out_sig = None
-    def __init__(self, process):
+    completion_list = set()
+
+    def __init__(self, parent, child=None):
         ''' Set generic parameters from Connection class '''
-        super(Signalroute, self).__init__(process, process)
+        super(Signalroute, self).__init__(parent, child or parent)
         self.parser = ogParser
-        self.completion_list = []
         self.blackbold = ()
         self.redbold = ()
         self.label_in = SignalList(parent=self)
         self.label_out = SignalList(parent=self)
         if not Signalroute.in_sig:
             # keep at class level as long as only one process is supported
-            # when copy-pasting a process the challel in/out signal lists
+            # when copy-pasting a process the channel in/out signal lists
             # are not parsed. Workaround is to keep the list "global"
             # to allow a copy of both process and channel
             # Needed for the image exporter, that copies the scene to a
             # temporary one
             Signalroute.in_sig = '{}'.format(',\n'.join(sig['name']
-                                   for sig in process.input_signals))
+                                   for sig in parent.input_signals))
             Signalroute.out_sig = '{}'.format(',\n'.join(sig['name']
-                                    for sig in process.output_signals))
+                                    for sig in parent.output_signals))
         self.label_in.setPlainText('[{}]'.format(self.in_sig))
         self.label_out.setPlainText('[{}]'.format(self.out_sig))
         self.label_in.document().contentsChanged.connect(self.change_siglist)
         self.label_out.document().contentsChanged.connect(self.change_siglist)
         for each in (self.label_in, self.label_out):
             each.show()
-        self.process = process
-        self.reshape()
+        #self.reshape()
 
     @Slot()
     def change_siglist(self):
         ''' Called when user modified the text of label_in or label_out '''
         for each in self.label_in, self.label_out:
             if not each.hasFocus():
-                if not unicode(each).startswith('['):
-                    each.setPlainText('[{}'.format(unicode(each)))
-                if not unicode(each).endswith(']'):
-                    each.setPlainText('{}]'.format(unicode(each)))
-        Signalroute.in_sig = unicode(self.label_in)[1:-1]
-        Signalroute.out_sig = unicode(self.label_out)[1:-1]
+                if not str(each).startswith('['):
+                    each.setPlainText('[{}'.format(str(each)))
+                if not str(each).endswith(']'):
+                    each.setPlainText('{}]'.format(str(each)))
+        Signalroute.in_sig = str(self.label_in)[1:-1]
+        Signalroute.out_sig = str(self.label_out)[1:-1]
 
 
     @property
     def start_point(self):
         ''' Compute connection origin - redefined function '''
-        parent_rect = self.process.boundingRect()
+        parent_rect = self.parent.boundingRect()
         return QPointF(parent_rect.x(), parent_rect.height() / 2)
 
     @property
@@ -319,9 +334,12 @@ class Signalroute(Connection):
         # Arrow always bumps at the screen edge
         try:
             view = self.scene().views()[0]
+            #view.update_phantom_rect()
+            # view_pos is the position of the view relative to the scene
             view_pos = view.mapToScene(
                            view.viewport().geometry()).boundingRect().topLeft()
             scene_pos_x = self.mapFromScene(view_pos).x()
+            #print view_pos.x(), scene_pos_x, view.sceneRect().x()
             return QPointF(scene_pos_x, self.start_point.y())
         except (IndexError, AttributeError):
             # In case there is no view (e.g. Export PNG from cmd line)
@@ -350,16 +368,111 @@ class Signalroute(Connection):
 
     def update_completion_list(self, pr_text):
         ''' Called after text has been edited '''
-        # TODO - call parseSingleElement, check sdlSymbols for examples
-        pass
+        from .sdlSymbols import CONTEXT
+        ast, _, _, _, _ = self.parser.parseSingleElement('signalroute',
+                                                         pr_text)
+        # ast is a dict:
+        # {'routes': [{'dest': str, 'source': str', 'signals': [str]}],
+        #  'name': u'c'} - dest and source can be 'ENV'
+        all_sigs = []
+        for each in ast['routes']:
+            source    = each['source']
+            dest      = each['dest']
+            sigs = [{'name': name} for name in each['signals']]
+            all_sigs.extend(sigs)
+            def get_sig(signal):
+                ''' If signal is declared, return the signature '''
+                for sig in CONTEXT.signals:
+                    if sig['name'].lower() == signal['name'].lower():
+                        return sig
+                return signal
+            if each['dest'] == 'ENV':
+                # output signals of process 'source'
+                process, = [p for p in CONTEXT.processes
+                            if p.processName.lower() == each['source'].lower()]
+                existing = [sig['name'].lower()
+                           for sig in process.output_signals]
+                for sig in sigs:
+                    if sig['name'].lower() not in existing:
+                        process.output_signals.append(get_sig(sig))
+            else:
+                # input signals of process 'source'
+                try:
+                    process, = [p for p in CONTEXT.processes
+                            if p.processName.lower() == each['dest'].lower()]
+                    existing = [sig['name'].lower()
+                            for sig in process.input_signals]
+                    for sig in sigs:
+                        if sig['name'].lower() not in existing:
+                            process.input_signals.append(get_sig(sig))
+                except ValueError:
+                    # No process found
+                    pass
+        existing = [sig['name'].lower() for sig in CONTEXT.signals]
+        for each in all_sigs:
+            if each['name'].lower() not in existing:
+                CONTEXT.signals.append(each)
 
     def resize_item(self, new_rect):
         ''' Called after signallist text has been edited '''
         pass
 
 
-class Controlpoint(QGraphicsPathItem, object):
+class Channel(Signalroute):
+    ''' Connection between two processes. Signalroute handles connections
+    between a function and the environment (edge of the screen). Here the
+    start, middle and end point are redefined. They are stored with
+    scene coordinates '''
+
+    @Slot(float, float)
+    def child_moved(self, delta_x, delta_y):
+        ''' When the connection child moves - redefined function '''
+        self._end_point.setX(self._end_point.x() - delta_x)
+        self._end_point.setY(self._end_point.y() - delta_y)
+
+    @property
+    def start_point(self):
+        ''' Compute connection origin - redefined function '''
+        return self._start_point
+
+    @start_point.setter
+    def start_point(self, scene_coord : QPointF):
+        ''' value is in scene coordinates '''
+        self._start_point = self.parent.mapFromScene(scene_coord)
+
+    @property
+    def end_point(self):
+        ''' Compute connection end point - redefined function '''
+        return self.parent.mapFromScene(self._end_point)
+
+    @end_point.setter
+    def end_point(self, scene_coord : QPointF):
+        ''' value is in scene coordinates '''
+        self._end_point = scene_coord
+
+    @property
+    def middle_points(self):
+        ''' Compute connection intermediate points - redefined function '''
+        for each in self._middle_points:
+            yield self.parent.mapFromScene(each)
+
+    @middle_points.setter
+    def middle_points(self, points_scene_coord):
+        self._middle_points = points_scene_coord
+
+    def add_point(self, scene_coord):
+        self._middle_points.append(scene_coord)
+
+    def paint(self, painter, _, ___):
+        ''' Apply anti-aliasing '''
+        painter.setRenderHint(QPainter.Antialiasing, True)
+        super(Channel, self).paint(painter, _, ___)
+
+
+
+class Controlpoint(QGraphicsPathItem):
     ''' Class handling one edge control point (to change bezier curves) '''
+    user_can_connect = False  # not used but will avoid exceptions
     def __init__(self, pos, edge):
         ''' Set the original control point - with color, shape '''
         path = QPainterPath()
@@ -445,7 +558,7 @@ class Edge(Connection):
         self.bezier = [self.mapFromScene(*self.edge['spline'][0])]
         # Bezier control points (groups of three points):
         assert(len(self.edge['spline']) % 3 == 1)
-        for i in xrange(1, len(self.edge['spline']), 3):
+        for i in range(1, len(self.edge['spline']), 3):
             self.bezier.append([Controlpoint(
                           self.mapFromScene(*self.edge['spline'][i + j]), self)
                           for j in range(3)])
@@ -493,7 +606,7 @@ class Edge(Connection):
         ''' On a mouse click, display the control points '''
         self.bezier_set_visible(True)
 
-  # pylint: disable=R0914
+    # pylint: disable=R0914
     def reshape(self):
         ''' Update the shape of the edge (redefined function) '''
         path = QPainterPath()
@@ -520,18 +633,30 @@ class Edge(Connection):
         try:
             # Add the transition label, if any (none for the START edge)
             font = QFont('arial', pointSize=8)
-            width = QFontMetrics(font).width(
-                    self.edge.get('label', 0))
+            metrics = QFontMetrics(font)
+            label = self.edge.get('label', '') or ''
+            lines = label.split('\n')
+            width = metrics.width(max(lines)) # longest line
+            height = metrics.height() * len(lines)
+            # lp is the position of the center of the text
             pos = self.mapFromScene(*self.edge['lp'])
-            #path.addText(pos.x() - width/2, pos.y(),
-            #        font, self.edge['label'])
             if not self.text_label:
                 self.text_label = QGraphicsTextItem(
                                  self.edge.get('label', ''), parent=self)
-            self.text_label.setX(pos.x() - width / 2)
-            self.text_label.setY(pos.y())
-            self.text_label.setFont(font)
-            self.text_label.show()
+                self.text_label.setX(pos.x() - width / 2)
+                self.text_label.setY(pos.y() - height / 2)
+                self.text_label.setFont(font)
+                # Make horizontal center alignment, as dot does
+                self.text_label.setTextWidth(
+                        self.text_label.boundingRect().width())
+                fmt = QTextBlockFormat()
+                fmt.setAlignment(Qt.AlignHCenter)
+                cursor = self.text_label.textCursor()
+                cursor.select(QTextCursor.Document)
+                cursor.mergeBlockFormat(fmt)
+                cursor.clearSelection()
+                self.text_label.setTextCursor(cursor)
+                self.text_label.show()
         except KeyError:
             # no label
             pass
